#include "libvdev/sglib.h"
#include "libvdev/util.h"
#include "libvdev/misc.h"
#include "libvdev/sbuf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdbool.h>
#include <mntent.h>
#include <sys/stat.h>

const char *progname = "vdev_modprobe";
const char *mtab = "/etc/mtab";
const char *lockfile = "/run/vdev/vdev_modprobe.lock";

static int fd = -1;
static struct flock fl;

typedef char* cstr;

// prototypes
SGLIB_DEFINE_VECTOR_PROTOTYPES (cstr);
SGLIB_DEFINE_VECTOR_FUNCTIONS (cstr);

struct sysfs_scan_context {
        char* uevent_path;
        struct sglib_cstr_vector* frontier;
};

// 'buf' must be freed after usage
static void get_sysfs_path(char **buf)
{
        struct mntent *e;
        FILE* fstab = NULL;
   
        fstab = setmntent(mtab, "r");
        if (!fstab) {
                vdev_error("%s: setmntent(): error trying to open /etc/mtab: '%s'\n", progname, strerror (errno));
                exit(EXIT_FAILURE);
        }
    
        *buf = (char*)malloc(sizeof(char) * 32);
        if (!*buf) {
                vdev_error ("%s: Memory allocation failure: '%s'\n", progname, strerror (errno));
                exit(EXIT_FAILURE);
        }
   
        *buf[0] = '\0';
         
        while ((e = getmntent(fstab))) {
                if (!strcmp(e->mnt_type, "sysfs")) {
                        sprintf( *buf, "%s", e->mnt_dir );
                        break;
                }
        }
   
        endmntent(fstab);
}

// scan a directory in /sys/devices directory, to find its child directories, that are pushed back to scan_ctx->frontier
// return 0 on success
// return -ENOMEM on OOM
// return -errno on failure to stat
static int scan_device_directory(char const* fp, void* cls)
{
        struct sysfs_scan_context *scan_ctx = (struct sysfs_scan_context*)cls;
   
        struct sglib_cstr_vector* frontier = scan_ctx->frontier;

        int rc = 0;  
        struct stat sb;
        char* fp_base = NULL;
        char* fp_dup = NULL;
   
        fp_base = rindex (fp, '/') + 1;
   
        if (fp_base == NULL)
                return 0;
   
        // skip . and .. 
        if (strcmp(fp_base, ".") == 0 || strcmp(fp_base, "..") == 0)
                return 0;
  
        // add directories
        rc = lstat (fp, &sb);
        if (rc != 0) {
                vdev_error ("%s: lstat('%s'): '%s'\n", progname, fp, strerror(errno));
                return -errno;
        }
   
        if (!S_ISDIR(sb.st_mode) && strcmp(fp_base, "uevent") != 0)
                return 0;
   
        fp_dup = vdev_strdup_or_null(fp);
        if (fp_dup == NULL)
                return -ENOMEM;
   
        if (S_ISDIR(sb.st_mode)) {
                rc = sglib_cstr_vector_push_back (frontier, fp_dup);
                if (rc != 0) {
                        free (fp_dup);
                        return rc;
                }
        /* this is a uevent; this directory is a device */      
        } else {
                scan_ctx->uevent_path = fp_dup;
        }
   
        return 0;
}

// get a uevent from a uevent file 
// replace newlines with '\0', making the uevent look like it came from the netlink socket
// (i.e. so it can be parsed by vdev_linux_parse_request)
// return 0 on success
// return -ENOMEM on OOM
// return -errno on failure to stat or read
static int vdev_read_uevent(char const* fp_uevent, char** ret_uevent_buf, size_t* ret_uevent_len)
{
        int rc = 0;
        struct stat sb;
        char* uevent_buf = NULL;
        size_t uevent_buf_len = 0;
        size_t uevent_len = 0;
   
        // get uevent size  
        rc = stat(fp_uevent, &sb);
        if (rc != 0) {
                vdev_error ("%s: stat('%s'): '%s'\n", progname, fp_uevent, strerror (errno));
                return -errno;
        }
        else
                uevent_buf_len = sb.st_size;
   
        // read the uevent as long as it's a regular file
        if (fp_uevent != NULL && (sb.st_mode & S_IFMT) == S_IFREG) {
                uevent_buf = VDEV_CALLOC(char, uevent_buf_len);
                if (uevent_buf == NULL)
                        return -ENOMEM;
      
                rc = vdev_read_file(fp_uevent, uevent_buf, uevent_buf_len);
                if (rc != 0) {
                        // failed in this 
                        vdev_error ("%s: read_file('%s'): '%s'\n", progname, fp_uevent, strerror (errno));
                        free(uevent_buf);
                }
                else {
                        for (unsigned int i = 0; i < uevent_buf_len; i++) {
                                if (uevent_buf[i] == '\n')               
                                        uevent_buf[i] = '\0';
                        }
         
                        // NOTE: the stat size is an upper-bound.  Find the exact number of bytes.
                        for (uevent_len = 0; uevent_len < uevent_buf_len; ) {
                                if (*(uevent_buf + uevent_len) == '\0')
                                        break;
                                uevent_len += strlen( uevent_buf + uevent_len ) + 1;
                        }
          
                        *ret_uevent_buf = uevent_buf;
                        *ret_uevent_len = uevent_len;
                }
        }
   
        return rc;
}

// make the full sysfs path from the dev path, plus an additional path 
// return NULL on OOM
static char* vdev_linux_sysfs_fullpath(char const* sysfs_mountpoint, char const* devpath, char const* attr_path)
{
        char* tmp = NULL;
        char* ret = NULL;
   
        tmp = vdev_fullpath(sysfs_mountpoint, devpath, NULL);
        if (!tmp)
                return NULL;
   
        ret = vdev_fullpath(tmp, attr_path, NULL);
        free(tmp);
   
        return ret;
}

// read the kernel-given device subsystem from sysfs 
// return 0 on success, and set *subsystem
// return -ENOMEM on OOM 
// return negative on readlink failure
static int vdev_linux_sysfs_read_subsystem(const char *mountpoint, char const* devpath, char** subsystem)
{
        int rc = 0;
        char linkpath[PATH_MAX+1];
        size_t linkpath_len = PATH_MAX;
        char* subsystem_path = NULL;
   
        memset(linkpath, 0, PATH_MAX+1);
   
        subsystem_path = vdev_linux_sysfs_fullpath(mountpoint, devpath, "subsystem");
        if (subsystem_path == NULL)
                return -ENOMEM;
   
        rc = readlink(subsystem_path, linkpath, linkpath_len);
        if (rc < 0) {
                vdev_error("%s: readlink('%s') %s\n", progname, subsystem_path, strerror(errno));
                free(subsystem_path);
                return -errno;
        }
   
        free(subsystem_path);
   
        *subsystem = vdev_basename(linkpath, NULL);
        if (*subsystem == NULL)
                return -ENOMEM;
   
        return 0;
}

static int modprobe(char const* fp_uevent)
{
        int rc = 0;
        struct stat sb;
        char* uevent_buf = NULL;
        size_t uevent_buf_len = 0;
        char* full_devpath = NULL;
   
        // get uevent
        rc = vdev_read_uevent(fp_uevent, &uevent_buf, &uevent_buf_len);
        if (rc != 0) {
                vdev_error ("%s: read_uevent('%s'): '%s'\n", progname, fp_uevent, strerror(errno));
                return rc;
        }
   
        if (uevent_buf_len == 0) {
                free(uevent_buf);
                return 0;
        }
   
        for (unsigned int i = 0; i < uevent_buf_len;) {
                char *tmp = rindex(uevent_buf + i, '=') + 1;
                if (!strncmp(uevent_buf+i, "MODALIAS=", strlen("MODALIAS=")) && strlen(tmp) > 0) {
                        FILE *pfin = NULL;
                        pid_t pid;
                        int wstatus;
                        char *tmp = rindex(uevent_buf + i, '=') + 1;
						if (strlen(tmp) > 2) {
							sbuf_t cmd;
							sbuf_init(&cmd);
							sbuf_concat(&cmd, 3, "/sbin/modprobe -b -q -- \"", tmp, "\"");
							pfin = epopen(cmd.buf, &pid);
							if (pfin) { 
                                fclose(pfin);
                                waitpid(pid, &wstatus, 0);
							}
							sbuf_free(&cmd);
					    }
                }
           
                i += strlen(uevent_buf + i) + 1;
        }
   
        // Try to create a lock file. Bear in mind that the system still may be read only at this point.
        // If so, we ignore the failed atempt because we'll try again when loading the next driver 
        if (fd == -1) {
                fd = open(lockfile, O_RDWR|O_CREAT, 0640); 
                if (fd < 0) {
                        switch(errno) {
                        case ENOMEM:
                                return -ENOMEM;
                        default:
                                return rc;
                        }      
                }
        } else {   
                return rc;
        }

        memset(&fl, 0, sizeof(fl));
   
        fl.l_type = F_WRLCK;     // Set a write lock
        fl.l_whence = SEEK_SET;  // Start at beginning of file
        fl.l_start = 0;          // Offset from beginning of file
        fl.l_len = 0;            // Lock entire file
   
        // F_SETLK(W) ignores it; F_OFD_SETLK(W) requires it to be zero
        fl.l_pid = 0;            //getpid(); // Set process ID
   
        // F_SETLKW specifies blocking mode
        fcntl(fd, F_SETLKW, &fl); // Set the lock
   
        return rc;
}

/**
 * \brief Variadic function
 */
static int find_devices_at_frontier(const char *sysfs_path, const char *device_frontier, struct sglib_cstr_vector *uevent_paths)
{
        int rc = 0;
        struct sglib_cstr_vector frontier;
        struct sysfs_scan_context scan_ctx;

        sglib_cstr_vector_init (&frontier);
   
        memset (&scan_ctx, 0, sizeof(struct sysfs_scan_context));
   
        scan_ctx.frontier = &frontier;

        rc = vdev_load_all (device_frontier, scan_device_directory, &scan_ctx);
        if (rc != 0) {
                vdev_error ("%s: vdev_load_all('%s'): '%s'\n", progname, device_frontier, strerror(errno));
                sglib_cstr_vector_free (&frontier);      
                return rc;
        }
   
        while (1) {
                unsigned long len = sglib_cstr_vector_size (&frontier);
                if (len == 0)
                        break;
   
                char* dir = sglib_cstr_vector_at (&frontier, len - 1);
                sglib_cstr_vector_set (&frontier, NULL, len - 1);
      
                sglib_cstr_vector_pop_back (&frontier);
      
                // scan for more devices 
                rc = vdev_load_all (dir, scan_device_directory, &scan_ctx);
                if (rc != 0) {
                        vdev_error ("%s: vdev_load_all('%s'): '%s'\n", progname, dir, strerror(errno));
                        free (dir);
                        break;
                }
      
                // is one of them a uevent?
                if (scan_ctx.uevent_path) {
                        const char *str;
                        bool is_ok = false;
         
                        char* uevent_path = vdev_strdup_or_null (scan_ctx.uevent_path);
                        if (!uevent_path) {
                                free (dir);
                                free (scan_ctx.uevent_path);
                                scan_ctx.uevent_path = NULL;
                                rc = -ENOMEM;
                                break;
                        }

                        rc = sglib_cstr_vector_push_back (uevent_paths, uevent_path);
                        if (rc < 0) {
                                ///free (uevent_path);  ???
                                free (dir);
                                free (scan_ctx.uevent_path);
                                scan_ctx.uevent_path = NULL;
                                break;
                        }
         
                        free (scan_ctx.uevent_path);
                        scan_ctx.uevent_path = NULL;
                }
      
                free (dir);
        }
   
        sglib_cstr_vector_free (&frontier);

        return rc;
}

static int find_devices(char *sysfs_mountpoint, struct sglib_cstr_vector* uevent_paths)
{
        int rc = 0;
        struct sglib_cstr_vector frontier;
        struct sysfs_scan_context scan_ctx;
   
        memset(&scan_ctx, 0, sizeof(struct sysfs_scan_context));
      
        char* devroot = NULL;

        sglib_cstr_vector_init (&frontier);
   
        scan_ctx.frontier = &frontier;
   
        devroot = vdev_fullpath (sysfs_mountpoint, "/devices", NULL);     // devroot <= /sys/devices
        if( devroot == NULL )      
                return -ENOMEM;
    
        rc = vdev_load_all (devroot, scan_device_directory, &scan_ctx);
        if( rc != 0 ) {
                vdev_error ("%s: vdev_load_all('%s'): '%s'\n", progname, devroot, strerror(errno)); 
                free (devroot);    
                sglib_cstr_vector_free (&frontier);      
                return rc;
        }
   
        free(scan_ctx.uevent_path);
        scan_ctx.uevent_path = NULL;
 
        for (unsigned long i = 0; i < sglib_cstr_vector_size (&frontier); i++)
                find_devices_at_frontier (sysfs_mountpoint, sglib_cstr_vector_at (&frontier, i), uevent_paths);
   
        sglib_cstr_vector_free (&frontier);
   
        return rc;
}

int main (int argc, char **argv)
{
        int rc, flag;
        struct sglib_cstr_vector uevent_paths;
        struct sglib_cstr_vector pci_paths;
        char sysfs_mountpoint[PATH_MAX+1];
        char *buf = NULL;
   
        if (access(mtab, F_OK) != 0) {
                snprintf (sysfs_mountpoint, PATH_MAX, "/sys"); 
        } else {
                get_sysfs_path (&buf);
                if (buf && buf[0] != '\0') {
                        snprintf (sysfs_mountpoint, PATH_MAX, buf);       
                        free (buf);   
                } else {
                        vdev_error ("%s: Cannot get sysfs path\n", progname);
                        exit (EXIT_FAILURE);
                }
        }
        
        send_to_background();

        sglib_cstr_vector_init (&uevent_paths);
        rc = find_devices (sysfs_mountpoint, &uevent_paths);
        
        sglib_cstr_vector_init (&pci_paths);
        
        for (unsigned long i = 0; i < sglib_cstr_vector_size(&uevent_paths); i++) {
                if (sysfs_get_parent_with_subsystem(sglib_cstr_vector_at(&uevent_paths, i), "pci")) {
                        sglib_cstr_vector_push_back(&pci_paths, sglib_cstr_vector_at (&uevent_paths, i));
                        sglib_cstr_vector_set(&uevent_paths, NULL, i);
                }
        }       
        sglib_cstr_vector_free (&uevent_paths);
      
        char *p[] = { "graphics", "pci", (char*)0 }; //"drm",
        for (unsigned i = 0; i < sizeof(p)/sizeof(char*) - 1; i++) {
                for (unsigned long j = 0; j < sglib_cstr_vector_size(&pci_paths); j++) {
                        char *subsystem = NULL;
                        char* full_devpath = NULL;
                        
                        // extract the devpath from the uevent path 
                        full_devpath = vdev_dirname(sglib_cstr_vector_at(&pci_paths, j), NULL);
                        if (full_devpath == NULL)
                                continue;

                        rc = vdev_linux_sysfs_read_subsystem("", full_devpath, &subsystem);
                        if (rc == 0) {
                                if (!strcmp(subsystem, p[i])) {
                                        if (!strcmp(subsystem, "pci")) {
                                                sbuf_t s;
                                                sbuf_init(&s);
                                                sbuf_concat(&s, 2, full_devpath, "/driver");
                                                if (access(s.buf, F_OK) == 0) {
                                                        /*
                                                        char buffer[PATH_MAX+1]={0};
                                                        if (readlink(s.buf, buffer, sizeof(buffer)) > 0) {
                                                                modprobe(sglib_cstr_vector_at(&pci_paths, j));
                                                                char *q[] = { "pcieport", (char*)0 }; 
                                                                char *tmp = rindex(buffer, '/')+1;
                                                                for (unsigned k = 0; k < sizeof(q)/sizeof(char*) - 1; k++) {
                                                                        if (!strcmp(tmp, q[k])) {                                     
                                                                                modprobe(sglib_cstr_vector_at(&pci_paths, j));
                                                                                break;
                                                                        }
                                                                }
                                                        }
                                                        * */
                                                } else {
                                                        modprobe(sglib_cstr_vector_at(&pci_paths, j));
                                                }
                                                sbuf_free(&s);
                                        } else {
                                                modprobe(sglib_cstr_vector_at(&pci_paths, j));
                                        }
                                        sglib_cstr_vector_set(&pci_paths, NULL, j);
                                }
                                free(subsystem);
                        }
                        free(full_devpath);
                }
        }
/*        
        for (unsigned long i = 0; i < sglib_cstr_vector_size(&pci_paths); i++) {
                if (sglib_cstr_vector_at(&pci_paths, i) == NULL)
                        continue;
                modprobe(sglib_cstr_vector_at(&pci_paths, i));
        }
*/
        sglib_cstr_vector_free (&pci_paths);
   
        if (fd > 0) {
                fl.l_type = F_UNLCK; // Release the lock
                fcntl(fd, F_SETLK, &fl); // Unlock the file
                close(fd);
                unlink(lockfile);
        } 
   	
        return 0;
}
