
#include "libvdev/sglib.h"
#include "libvdev/util.h"
#include "libvdev/misc.h"
#include "libvdev/sbuf.h"

#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <mntent.h>

const char *progname = "vdev_modprobe_wait";
const char *mtab = "/etc/mtab";

typedef char* cstr;

// prototypes
SGLIB_DEFINE_VECTOR_PROTOTYPES (cstr);
SGLIB_DEFINE_VECTOR_FUNCTIONS (cstr);

static void get_mounted_devices(struct sglib_cstr_vector *mounted_devices)
{
        DIR *dirp;
        struct dirent *E;
        struct mntent *e;
        FILE* fstab = NULL;   
        struct sglib_cstr_vector devices;
        const char *dir = "/sys/block";
        
        sglib_cstr_vector_init (&devices);
        
        dirp = opendir(dir);
        if (!dirp) {
                vdev_error ("%s error opening current directory: %s\n", progname, strerror(errno));
                sglib_cstr_vector_free(&devices);
                sglib_cstr_vector_free(mounted_devices);
                exit(EXIT_FAILURE);
        };

        while ((errno=0, E=readdir(dirp))) {
                if (!strstr(E->d_name, ".") && !strstr(E->d_name, "..")) {
                        sglib_cstr_vector_push_back(&devices, vdev_strdup_or_null(E->d_name));
                }
        }
        if (errno) {
                vdev_error ("%s error reading current directory: %s\n", progname, strerror(errno));
                closedir(dirp);
                sglib_cstr_vector_free(&devices);
                sglib_cstr_vector_free(mounted_devices);
                exit(EXIT_FAILURE);
        }
        closedir(dirp);
   
        fstab = setmntent(mtab, "r");
        if (!fstab) {
                vdev_error("%s: setmntent(): error trying to open /etc/mtab: '%s'\n", progname, strerror (errno));
                sglib_cstr_vector_free(&devices);
                exit(EXIT_FAILURE);
        }
         
        while ((e = getmntent(fstab))) {
                if (!strncmp(e->mnt_fsname, "/dev/", 5)) {
                        char *str = vdev_strdup_or_null(strstr(e->mnt_fsname, "/dev/")+5);
                        for (int i = 0; i < sglib_cstr_vector_size(&devices); i++) {
                                char *aux = vdev_strdup_or_null(sglib_cstr_vector_at(&devices, i));
                                if (!strncmp(str, aux, strlen(aux))) {
                                        bool already_exists = false;
                                        for (int j = 0; j < sglib_cstr_vector_size(mounted_devices); j++) {
                                                if (!strcmp(aux, sglib_cstr_vector_at(mounted_devices, j))) {
                                                        already_exists = true;
                                                        break;
                                                }
                                        }
                                        if (!already_exists) {
                                                int rc = sglib_cstr_vector_push_back(mounted_devices, aux);
                                                if (rc < 0) {
                                                        endmntent(fstab);
                                                        sglib_cstr_vector_free(&devices);
                                                        sglib_cstr_vector_free(mounted_devices);
                                                        exit(EXIT_FAILURE);
                                                }
                                        }
                                }
                        }
                }
        }
        endmntent(fstab);
        sglib_cstr_vector_free(&devices);
}

int main(int argc, char *argv[])
{
        DIR *dirp;
        struct dirent *E;   
        struct sglib_cstr_vector block_devices;
        struct sglib_cstr_vector usb_devices;
        struct sglib_cstr_vector mounted_devices;
        const char *blockdir = "/sys/dev/block";
        const char *usbdir = "/sys/bus/usb/drivers/usb";
        sbuf_t pid1uid;
        bool user_session_started = false;                
        
        sbuf_init(&pid1uid);
        while (!user_session_started) {
                dirp = opendir("/proc");
                if (!dirp) {
                        vdev_error("%s: opendir(): %s\n", progname, strerror(errno));
                        sbuf_free(&pid1uid);
                        exit(EXIT_FAILURE);
                }
	
                while ((E = readdir(dirp))) {
                        sbuf_t s;
                        FILE *fp = NULL;
                        char line[1024];
                        if (!isdigit(E->d_name[0]))
                                continue;
                        sbuf_init(&s);
                        sbuf_concat(&s, 3, "/proc/", E->d_name, "/loginuid");
                        fp = fopen(s.buf, "r");
                        if (!fp) {
                                vdev_error("%s: fopen(%s): %s\n", progname, s.buf, strerror(errno));
                                sbuf_free(&s);
                                continue;
                        }
                        sbuf_free(&s);
                        if (fgets(line, 1024, fp)) {
                                line[strcspn(line, "\n")] = '\0';
                                if (pid1uid.len == 0)
                                        sbuf_addstr(&pid1uid, line);
                                else if (strcmp(line, pid1uid.buf)) {
                                        user_session_started = true;
                                        fclose(fp);
                                        break;
                                }
                        }
                        fclose(fp);
                }	
                closedir(dirp);
	}
        sbuf_free(&pid1uid);
        
        sglib_cstr_vector_init (&mounted_devices);
   
        if (access(mtab, F_OK) != 0) {
                sglib_cstr_vector_free(&mounted_devices);
                exit(EXIT_FAILURE); 
        }
        else get_mounted_devices(&mounted_devices);
        
        send_to_background();
        
        dirp = opendir(blockdir);
        if (!dirp) {
                vdev_error ("%s error opening current directory: %s\n", progname, strerror(errno));
                exit(EXIT_FAILURE);
        };

        sglib_cstr_vector_init (&block_devices);
        while ((errno=0, E=readdir(dirp))) {
                if (!strstr(E->d_name, ".") && !strstr(E->d_name, "..")) {
                        sglib_cstr_vector_push_back(&block_devices, vdev_strdup_or_null(E->d_name));
                }
        }
        if (errno) {
                vdev_error ("%s error reading current directory: %s\n", progname, strerror(errno));
                closedir(dirp);
                sglib_cstr_vector_free(&mounted_devices);
                sglib_cstr_vector_free(&block_devices);
                exit(EXIT_FAILURE);
        }
        closedir(dirp);

        sglib_cstr_vector_init (&usb_devices);

        for (int i = 0; i < sglib_cstr_vector_size(&block_devices); i++) {
                sbuf_t sl;
                char buffer[PATH_MAX+1]={0};              
                char *token = NULL;
                char *tmp = NULL;
                bool is_mounted = false;
                
                sbuf_init(&sl);
                sbuf_concat(&sl, 3, blockdir, "/", sglib_cstr_vector_at(&block_devices, i));
                
                if (readlink(sl.buf, buffer, sizeof(buffer)) < 0) { 
                        vdev_error("readlink(%s) error", sl.buf);
                        sbuf_free(&sl);
                        sglib_cstr_vector_free(&block_devices);
                        sglib_cstr_vector_free(&mounted_devices);
                        sglib_cstr_vector_free(&usb_devices);
                        exit(EXIT_FAILURE);
                }
                
                tmp = rindex(buffer, '/')+1;
                if (!strncmp(tmp, "loop", 4) && isdigit(tmp[4])) {
                        sbuf_free(&sl);
                        continue;
                }
                
                for (int j = 0; j < sglib_cstr_vector_size(&mounted_devices); j++) {
                        const char *str = sglib_cstr_vector_at(&mounted_devices, j);
                        if (!strncmp(tmp, str, strlen(str))) {
                                 is_mounted = true;
                                 break; 
                        }
                }
                if (is_mounted) {
                        sbuf_free(&sl);
                        continue;
                }
                token = strtok(strstr(buffer, "/devices/"), "/");
                while (token != NULL) {
                        bool already_exists = false;
                        if (isdigit(token[0])) {
                                for (int j = 0; j < sglib_cstr_vector_size(&usb_devices); j++) {
                                        if (!strcmp(token, sglib_cstr_vector_at(&usb_devices, j))) {
                                                already_exists = true;
                                                break;
                                        }
                                }
                                if (!already_exists)
                                        sglib_cstr_vector_push_back(&usb_devices, vdev_strdup_or_null(token));
                        }
                        token = strtok(NULL, "/");
                }
                sbuf_free(&sl);
        }
        
        dirp = opendir(usbdir);
        if (!dirp) {
                vdev_error ("%s error opening current directory: %s\n", progname, strerror(errno));
                sglib_cstr_vector_free(&block_devices);
                sglib_cstr_vector_free(&mounted_devices);
                sglib_cstr_vector_free(&usb_devices);
                exit(EXIT_FAILURE);
        };
 
        while ((errno=0, E=readdir(dirp))) {
                if (isdigit(E->d_name[0]) && !strstr(E->d_name, ".") && !strstr(E->d_name, "..")) {
                        for (int i = 0; i < sglib_cstr_vector_size(&usb_devices); i++) {
                                if (!strcmp(E->d_name, sglib_cstr_vector_at(&usb_devices, i))) {
                                        FILE *pfin = NULL;
                                        pid_t pid;
                                        int wstatus;
                                        sbuf_t cmd;
                                        sbuf_init(&cmd);
                                        sbuf_concat(&cmd, 3, "/bin/echo ", E->d_name, " |sudo tee /sys/bus/usb/drivers/usb/unbind >/dev/null");
                                        pfin = epopen(cmd.buf, &pid);
                                        if (pfin) { 
                                                fclose(pfin);
                                                waitpid(pid, &wstatus, 0);
                                                sbuf_init(&cmd);
                                                sbuf_concat(&cmd, 3, "/bin/echo ", E->d_name, " |sudo tee /sys/bus/usb/drivers/usb/bind >/dev/null");
                                                pfin = epopen(cmd.buf, &pid);
                                                if (pfin) { 
                                                        fclose(pfin);
                                                        waitpid(pid, &wstatus, 0);
                                                }
                                        }
                                        sbuf_free(&cmd);
                                }
                        }
                }
        }

        sglib_cstr_vector_free(&block_devices);        
        sglib_cstr_vector_free(&usb_devices);
        sglib_cstr_vector_free(&mounted_devices);
 
        return 0;
}
